# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

import re
from functools import reduce

"""
The Return value of a Tcl command or procedure is basically a string. Without
knowledge about the structure or type of the return value, it is not possible to
convert this string to a meaningful data structure such as a list or dict in Python.

Return values from Tcl are wrapped in *TclRemoteObjRef* objects. Passing a
*TclRemoteObjRef* back to Tcl via *TclToolInterface* method calls causes the
interpreter to use a reference that was maintained within the Tcl tool as
argument, instead of passing a newly created string as argument. While this
should not make a difference *in theory*, some Tcl-based tools rely on
internal representations or object addresses of opaque handles to stay the same.
A *TclRemoteObjRef* can be converted to a string via *str()*.
"""

class TclRemoteObjRef:
    """
    Tcl return values are encapsulated in instances of TclRemoteObjRef.
    Using str(), the string representation of the TclRemoteObjRef can be
    accessed. 
    """

    def __init__(self, tool, cmd_idx:int, value:str, cmd:str):
        self.tool = tool
        self.cmd_idx = cmd_idx
        self.value = value
        self.cmd=cmd

    def __repr__(self):
        return f'Tcl{repr(self.value)}'

    def ref_str(self):
        """
        Returns:
            A Tcl string referencing the corresponding Tcl object in the
            $cmd_results Tcl array, e. g. '$cmd_results(123)'.
        """
        return f"$cmd_results({self.cmd_idx})"

    def __str__(self):
        return self.value

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __getattr__(self, name):
        """
        Similar to TclToolInterface.__getattr__. In addition to methods of
        TclToolInterface, methods returned by this __getattr__ include the
        TclRemoteObjRef as part of the Tcl string that is to be evaluated.

        Exmaple::
        
            v = t.concat("MyObj")
            v.mymethod("arg1", "arg2")

        If TclTool.called_object_pos is 'first', the object name is used as Tcl
        command name: 'MyObj mymethod arg1 arg2'

        If TclTool.called_object_pos is 'second', the method name is the command
        name and the object name comes right after the method name:
        'mymethod MyObj arg1 arg2'

        If TclTool.called_object_pos is 'last', the object name is appended
        at the end of the argument list: 'mymethod arg1 arg2 MyObj'
        """

        return lambda *args, **kwargs: self.proc_call(name, args, kwargs)

    def proc_call(self, name, args, kwargs):
        """
        Wrapper for TclTool.proc_call for using TclRemoteObjRefs in an
        object-oriented way. Internally invoked; do not use directly.
        """
        pos = self.tool.called_object_pos
        if pos == "first":
            args = (name, ) + args
            name = self.ref_str()
        elif pos == "second":
            args = (self, ) + args
        elif pos == "last":
            args = args + (self, )
        else:
            raise ValueError('called_object_pos must be "first", "second" or "last".')
        return self.tool.proc_call(name, args, kwargs)

# TODO: Is there a problem with escape_braces?
def escape_braces(data) -> str:
    return re.sub(r"([{}])", r"\\\1", data)

def encode(obj, nested:bool=False) -> str:
    """
    Encodes Python lists, tuples, dictionaries and simple scalar types in
    Tcl-friendly strings. 

    Returns:
        obj encoded as Tcl string
    """

    if isinstance(obj, (list, tuple)):
        return "{" + " ".join(map(lambda elem: encode(elem, nested=True), obj)) + "}"
    elif isinstance(obj, dict):
        item_list = reduce(lambda a, b: a+b, obj.items())
        return "{" + " ".join(map(lambda elem: encode(elem, nested=True), item_list)) + "}"
    elif isinstance(obj, TclRemoteObjRef) and not nested:
        return obj.ref_str()
    else:
        return "{" + escape_braces(str(obj)) + "}"



# SPDX-FileCopyrightText: 2025 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

from importlib.metadata import version, PackageNotFoundError

try:
    version = version("notcl")
except PackageNotFoundError:
    version = "unknown"



# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

from .message import Message


class TclHello(Message):
    __slots__ = () # TODO: 
    keys_required = ["patchlevel", "commands", "globals", "nameofexecutable"]

class PyProcedureCall(Message):
    __slots__ = ()
    keys_required = ["command"]

class TclProcedureResult(Message):
    __slots__ = ()
    keys_required = ["err_code", "result", "cmd_idx"]

class PyExit(Message):
    __slots__ = ()
    keys_required = ["quit"]



# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

from .. import TclTool

class Yosys(TclTool):
    def cmdline(self):
        return ["yosys", "-c", self.script_name()]


# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

from .message import RawMessage, Message, WrongMessageClass
from . import msg_classes as msg

class BridgeClient:
    """This is a dummy client for testing BridgeServer.
    In production, this client is implemented in Tcl."""
    def __init__(self, fn_tcl2py, fn_py2tcl):
        self.fn_tcl2py = fn_tcl2py
        self.fn_py2tcl = fn_py2tcl

    def send(self, msg: Message):
        with open(self.fn_tcl2py, "wb") as f_out:
            msg.to_raw_message().send_to_pipe(f_out)


    def recv(self, permitted_msg_classes):
        """permitted_msg_classes can be either one message class or a list of message classes."""

        with open(self.fn_py2tcl, "rb") as f_in:
            raw_message = RawMessage.from_pipe(f_in)
        return raw_message.to_message(permitted_msg_classes)

    def run(self):
        self.send(msg.TclHello(nameofexecutable="dummy_client"))

        while True:
            m_recv = self.recv([msg.PyProcedureCall, msg.PyExit]) 
            if isinstance(m_recv, msg.PyExit):
                break
            elif isinstance(m_recv, msg.PyProcedureCall):
                result = m_recv.command.upper()
                self.send(msg.TclProcedureResult(result=result, err_code="0"))
            else:
                raise Exception(f"Unexpected message: {m_recv}")



# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

import pytest
import subprocess
from .. import TclTool, TclError

class Tclsh(TclTool):
    def cmdline(self):
        return ["tclsh", self.script_name()]

def test_set_string():
    with Tclsh() as t:
        t.set("myvar", "ThisIsATest")
        v=t.set("myvar")
        assert str(v) == "ThisIsATest"

def test_set_int():
    with Tclsh() as t:
        t.set("myvar", 123459)
        v=t.set("myvar")
        assert int(v) == 123459

def test_set_float():
    with Tclsh() as t:
        t.set("myvar", 3.14)
        v=t.set("myvar")
        assert abs(float(v) - 3.14) < 0.01 


def test_expr():
    with Tclsh() as t:
        v=t.expr(9, '+', 3, '*', 11)
        assert int(v) == 9+3*11

def test_list_of_strings():
    with Tclsh() as t:
        v=t.lreverse(["one", "two", "three"])
        assert str(v).split(" ") == ["three", "two", "one"]

def test_list_of_ints():
    with Tclsh() as t:
        v=t.lreverse([5,6,7])
        assert list(map(int, str(v).split(" "))) == [7,6,5]


def test_dict():
    with Tclsh() as t:
        v=t.dict("merge", {"key1":"value1"}, {"key2":"value2", "key3":"value3"})
        assert str(v).split(" ") == ["key1", "value1", "key2", "value2", "key3", "value3"]

def test_expr_log_commands_fancy(capsys):
    with Tclsh(log_commands=True, log_fancy=True) as t:
        v=t.expr(9, '+', 3, '*', 11)
        assert int(v) == 9+3*11
    cap = capsys.readouterr()
    assert cap.out == "\x1b[93m[notcl]\x1b[0m Cmd:\x1b[0m expr {9} {+} {3} {*} {11}\n"
    assert cap.err == ""

def test_expr_log_commands_plain(capsys):
    with Tclsh(log_commands=True, log_fancy=False) as t:
        v=t.expr(9, '+', 3, '*', 11)
        assert int(v) == 9+3*11
    cap = capsys.readouterr()
    assert cap.out == "[notcl] Cmd: expr {9} {+} {3} {*} {11}\n"
    assert cap.err == ""

def test_expr_log_commands_disabled(capsys):
    with Tclsh(log_commands=False) as t:
        v=t.expr(9, '+', 3, '*', 11)
        assert int(v) == 9+3*11
    cap = capsys.readouterr()
    assert cap.out == ""
    assert cap.err == ""

def test_explicit_ref():
    with Tclsh() as t:
        v1=t("expr 44 - 2")
        v2=t(f"expr {v1.ref_str()} / 2")
        assert int(v2) == 21
        v3=t(f"format Hello%i {v2.ref_str()}")
        assert str(v3) == "Hello21"

def test_implicit_ref():
    with Tclsh() as t:
        v1=t.expr(44, '-', 2)
        v2=t.expr(v1, '/', 2) 
        assert int(v2) == 21
        v3=t.format("Hello%i", v2)
        assert str(v3) == "Hello21"

def test_tclerror():
    with Tclsh() as t:
        with pytest.raises(TclError):
            t.expr("*", "+")

        with pytest.raises(TclError):
            t("does_not_exist")

        v=t.expr(1, '+', 1)
        assert int(v)==2

def test_lists():
    with Tclsh() as t:
        ret = t.list(1,2,3)
        assert str(ret) == '1 2 3'

def test_boolargs():
    with Tclsh() as t:
        ret = t.list(myarg=True)
        assert str(ret) == '-myarg'

        ret = t.list(myarg=False)
        assert str(ret) == ''

def test_repr():
    with Tclsh() as t:
        ret = t.list("hello", "world")
        assert repr(ret) == "Tcl'hello world'"

def test_child_terminates_nonzero():
    reaches1 = False
    reaches2 = False
    with pytest.raises(subprocess.CalledProcessError):
        with Tclsh() as t:
            reaches1 = True
            v=t.exit(1)
            reaches2 = True

    assert reaches1
    assert not reaches2

def test_child_terminates_zero():
    reaches1 = False
    reaches2 = False
    with pytest.raises(subprocess.CalledProcessError):
        with Tclsh() as t:
            reaches1 = True
            v=t.exit(0)
            reaches2 = True

    assert reaches1
    assert not reaches2

def test_context_forwards_excepts():
    with pytest.raises(ValueError):
        with Tclsh() as t:
            raise ValueError("test")  


# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

import tempfile
from contextlib import contextmanager
from pathlib import Path
from enum import Enum
import os

from .message import RawMessage, Message, WrongMessageClass


class BridgeServer:
    class State(Enum):
        NotListening = 0
        WaitForRecv = 1
        WaitForSend = 2

    def log(self, message):
        """Replace this method with something else to change the log mechanism"""
        print(f"[BridgeServer] {message}")
    
    def __init__(self, custom_log_func=None):
        self.cm = None
        self.state = self.State.NotListening
        if custom_log_func:
            self.log = custom_log_func

    @contextmanager
    def contextmanager(self):
        assert self.state == self.State.NotListening
        with tempfile.TemporaryDirectory() as tmp_dir:
            self.fn_tcl2py = str(Path(tmp_dir) / "tcl2py")
            self.fn_py2tcl = str(Path(tmp_dir) / "py2tcl")
            os.mkfifo(self.fn_tcl2py)
            os.mkfifo(self.fn_py2tcl)
            self.state = self.state.WaitForRecv
            try:
                yield self
            finally:
                self.log("BridgeServer: connection closed.")
                self.state = self.State.NotListening

    def recv_raw(self) -> RawMessage:
        assert self.state == self.State.WaitForRecv
        self.log(f"BridgeServer: Opening tcl2py pipe {self.fn_py2tcl} to receive message...")
        with open(self.fn_tcl2py, "rb") as f_in:
            m_recv = RawMessage.from_pipe(f_in)
        self.log(f"BridgeServer: tcl2py pipe {self.fn_py2tcl} closed.")
        self.state = self.State.WaitForSend
        return m_recv

    def send_raw(self, send_data: RawMessage):
        assert self.state == self.State.WaitForSend
        self.log(f"BridgeServer: Opening py2tcl pipe {self.fn_py2tcl} to send message...")
        with open(self.fn_py2tcl, "wb") as f_out:
            send_data.send_to_pipe(f_out)
        self.log(f"BridgeServer: py2tcl pipe {self.fn_py2tcl} closed.")
            
        self.state = self.State.WaitForRecv

    def recv(self, permitted_msg_classes) -> Message:
        """permitted_msg_classes can be either one message class or a list of message classes."""
        return self.recv_raw().to_message(permitted_msg_classes)

    def send(self, message: Message):
        raw_msg = message.to_raw_message()
        self.send_raw(raw_msg)

    def __enter__(self):
        assert self.cm == None
        self.cm = self.contextmanager()
        return self.cm.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        assert self.cm
        cm = self.cm
        self.cm = None
        return cm.__exit__(exc_type, exc_value, traceback)


# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

from ..bridge_server import BridgeServer
from ..bridge_client import BridgeClient

import threading
from .. import msg_classes as msg
from contextlib import contextmanager

@contextmanager
def bridge_server_with_client():
    with BridgeServer() as bs:
        client = BridgeClient(fn_tcl2py=bs.fn_tcl2py, fn_py2tcl=bs.fn_py2tcl)
        client_thread = threading.Thread(target=client.run)
        client_thread.start()

        try:
            yield bs
        finally:
            client_thread.join()

def test_bridge_server_client():
    with bridge_server_with_client() as bs:
        r=bs.recv(msg.TclHello)
        print(f"received {r}")
        
        for cmd in ["abc", "Defgh", "Guten Tag"]:
            bs.send(msg.PyProcedureCall(command=cmd))
            resp=bs.recv(msg.TclProcedureResult)
            print(f"received {resp}")
            assert resp.result == cmd.upper()
        bs.send(msg.PyExit())



# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

from .. import TclTool
from pathlib import Path
import os

class Vivado(TclTool):
    def cmdline(self):
        return ["vivado",
            "-mode", "tcl",
            "-nojournal",
            "-source", self.script_name()
        ]

    @staticmethod    
    def vivado_dir() -> Path:
        """
        Returns the XILINX_VIVADO Path, e.g. for locating Verilog files of
        the Xilinx cell libraries.
        """
        return Path(os.environ["XILINX_VIVADO"])


# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

import os
import traceback
import subprocess
import importlib.resources
import signal
from typing import Literal, Union, Optional
from abc import ABC, abstractmethod
from pathlib import Path
from contextlib import contextmanager

from . import tclobj, tcl
from . import msg_classes as msg

from .tclobj import TclRemoteObjRef
from .bridge_server import BridgeServer

class ANSITerm:
    FgBrightYellow = "\x1b[93m"
    BgRed = "\x1b[41m"
    BgGreen = "\x1b[42m"
    Reset = "\x1b[0m"

class TclError(Exception):
    """
    Errors in Tcl are forwarded to Python and raised in the form of a TclError
    exception. TclErrors can be caught and handled in Python. The corresponding
    TclTool remains usable.
    """
    def __init__(self, text: str):
        super().__init__()
        self.text = text

    def __str__(self):
        return self.text

class ChildProcessExited(Exception):
    """
    This exception is raised on abnormal / early termination of the Tcl child
    process through TclTool._sigchld_handler. It interrupts the
    current TclTool context. It is caught in TclTool.contextmanager.
    Typically, TclTool.contextmanager will then raise a subprocess.CalledProcessError
    based on the child's return code.
    """
    pass


class TclTool(ABC):
    called_object_pos = "second"
    """
    Used in TclRemoteObjRef.proc_call, specifying the invokation style for
    TclRemoteObjRef methods. Can be overwritten in subclasses.
    Must be 'first', 'second' oder 'last' (see TclRemoteObjRef.proc_call). 
    """ 

    def __init__(self, cwd:Optional[Union[Path, str]]=None, interact:bool=False,
            log_commands:bool=True, log_retvals:bool=False, log_fancy:bool=True,
            debug_tcl:bool=False, debug_py:bool=False, abort_on_error:bool=True):
        """
        Args:
            cwd: Directory in which to Tcl-based tool is run.
            interact: If set to True, the Tcl-based tool will not be terminated
                at the end of the with block. Instead, the Tcl tool remains open
                for the user to interact with. In this case, the Python script
                will wait for the user to close the tool manually before continuing.
            logs: Enables printing all commands sent to the Tcl tool.
            log_retvals: Enables printing of all return values received from Tcl tool.
            log_fancy: Enables colorful printing of commands, return values and
                errors using ANSI escape sequences.
            debug_tcl: Enable detailed debug output by the Tcl script running in
                the Tcl tool.
            debug_py: Enable detailed debug output for Python side.
            abort_on_error: Terminate child process when a Tcl error occured, even
                if interact is set to True.
        """

        self.interact = interact
        self.log_commands = log_commands
        self.log_retvals = log_retvals
        self.log_fancy = log_fancy
        self.debug_tcl=debug_tcl
        self.debug_py=debug_py
        self.abort_on_error = abort_on_error
        self._script_name = None
        self.cm = None
        
        if not cwd:
            self.cwd = Path.cwd()
        else:
            self.cwd = Path(cwd)

    @abstractmethod
    def cmdline(self) -> list[str]:
        """
        Returns:
            A list of strings that is passed to subprocess.Popen for invoking
            the Tcl tool. The first list element is the program name or full path
            to the executable. The subsequent list of arguments should be set
            in such a way that the Tcl script returned by self.script_name() is
            executed at startup of the invoked Tcl tool.
        """ 
        pass

    def script_name(self) -> str:
        """
        Returns:
            Filename of notcl.tcl file. Outside of the TclTool.mdline() call
            by TclTool.contextmanager, a placeholder string is retuned.
        """
        # This used to be a call to pkg_resources. After migration to
        # importlib.resources, this now is only a wratter to get a value
        # passed from contextmanager(self), which obtains it as context
        # from importlib.resources.as_file.
        if self._script_name==None:
            return "<placeholder for notcl.tcl>"
        else:
            return self._script_name

    def env(self) -> dict[str, str]:
        """
        Returns:
            Modified environment for child process.
        """
        env = os.environ.copy()
        env["NOTCL_PIPE_TCL2PY"]=self.bs.fn_tcl2py
        env["NOTCL_PIPE_PY2TCL"]=self.bs.fn_py2tcl
        env["NOTCL_DEBUG_TCL"]=("1" if self.debug_tcl else "0")
        return env

    def debug_log(self, message:str):
        """
        Prints debug message, if debug output is enabled by debug_py flag. 
        """
        if self.debug_py:
            print(f"[notcl] Python: {message}")

    def _sigchld_handler(self, sig, frame):
        """
        Called when child process exists, vis SIGHCLD Unix signal.
        This will typically interrupt the open() call in BridgeServer.recv_raw.
        """
        self.debug_log("Received SIGCHLD (child process terminated)")
        raise ChildProcessExited()
            

    @contextmanager
    def contextmanager(self):
        """
        Coroutine for invoking the Tcl tool on entry of the with block and
        waiting for the Tcl tool to finish when leaving the with block. It is
        invoked and managed internally by __enter__ and __exit__;
        not normally used externally.
        """

        with BridgeServer(custom_log_func=self.debug_log) as self.bs,\
            importlib.resources.as_file(importlib.resources.files(tcl).joinpath("notcl.tcl")) as script_name:

            try:
                self._script_name = script_name
                cmdline = self.cmdline()
            finally:
                # Clear self._script_name to make sure it never contains an invalid
                # invalid path (after leaving the script_name context here):
                self._script_name = None
            
            env = self.env()
            cwd = self.cwd
            
            signal.signal(signal.SIGCHLD, self._sigchld_handler)
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            
            self.proc = subprocess.Popen(cmdline, cwd=cwd, env=env)
            
            clean_exit = True

            if self.interact:
                quit = '0'
            else:
                quit = '1'

            try:
                self.hello=self.bs.recv(msg.TclHello)
                self.debug_log(f"Received TclHello: {self.hello}")
                try:
                    yield TclToolInterface(self)
                except ChildProcessExited:
                    clean_exit = False
                except:
                    self.debug_log("Caught exception in TclTool context.")
                    if self.abort_on_error:
                        quit = '1'
                    else:
                        s = traceback.format_exc()
                        self.log("error",
                            "Following exception is held back and will be raised "
                            f"once the Tcl child process exists:\n{s}")
                    raise
            
                finally:
                    if clean_exit:
                        self.debug_log("Sending PyExit")
                        
                        if quit == '0':
                            self.log("info", "Python control finished. Please exit Tcl tool to continue Python script.")
                        self.bs.send(msg.PyExit(quit=quit))
            except:    
                raise
            finally:
                # TODO: There is at least one unhandled race condition here:
                # What if we receive SIGCHLD before the signal handler is disabled?

                signal.signal(signal.SIGCHLD, signal.SIG_IGN)
                signal.signal(signal.SIGINT, signal.SIG_DFL)
                
                self.debug_log("TclTool context finished, waiting for child process to terminate.")
                rc = self.proc.wait()

                if rc==0 and clean_exit:
                    self.debug_log("Child process terminated with return code {rc}.")
                else:
                    self.debug_log("Child process terminated with abnormal return code {rc}.")
                    # If clean_exit is False, CalledProcessError is raised even though rc might be 0.
                    raise subprocess.CalledProcessError(rc, cmdline)

    def __enter__(self):
        assert self.cm == None
        self.cm = self.contextmanager()
        return self.cm.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        assert self.cm
        cm = self.cm
        self.cm = None
        return cm.__exit__(exc_type, exc_value, traceback)

    def proc_call(self, cmd:str, args:list, kwargs:dict):
        """
        Calls a procedure or command within Tcl. This method is invoked
        internally by calling a TclToolInterface method; do not use directly.

        Args:
            cmd: name of the Tcl procedure or command.
            args: list of positional arguments.
            kwargs: dictionary of keyword arguments.
        """

        # Reason why we have args and kwargs here instead of *args and **kwards: This
        # way we can support -self or -cmd als names Tcl command arguments.

        full_cmd = [cmd]
        for k, v in kwargs.items():
            if type(v)==bool:
                if v:
                    full_cmd.append("-"+k)
            else:
                full_cmd.append("-"+k)
                full_cmd.append(tclobj.encode(v))
        
        for arg in args:
            full_cmd.append(tclobj.encode(arg))

        return self.eval(" ".join(full_cmd))

    def log(self, log_type:Literal["command", "info", "retval", "error"], data:str):
        """
        log prints log messages for every Tcl command invokation and every
        received return value or error. It is enabled/disabled by the logs
        argument passed to __init__.
        """
        assert log_type in ("command", "retval", "error", "info")
        if log_type == "command":
            self.debug_log(f"Running command: {data}")
            if not self.log_commands:
                return
            log_symbol = "Cmd:"
        elif log_type == "retval":
            self.debug_log(f"Return value: {data}")
            if not self.log_retvals:
                return
            log_symbol = "Result:"
        elif log_type == "error":
            self.debug_log(f"Received error as return value: {data}")
            log_symbol = "Error:"
        elif log_type == "info":
            log_symbol = "Info:"

        style_notcl = ""
        style_reset = ""
        style_symbol = ""

        if self.log_fancy:
            style_notcl = ANSITerm.FgBrightYellow
            style_reset = ANSITerm.Reset
            if log_type=="error":
                style_symbol = ANSITerm.BgRed
            elif log_type=="info":
                style_symbol = ANSITerm.BgGreen
        
        print(f"{style_notcl}[notcl]{style_reset} {style_symbol}{log_symbol}{style_reset} {data}")

    def eval(self, cmd: str) -> TclRemoteObjRef:
        """
        Low-level method that passes a string to Tcl for evaluation.
        """
        self.log("command", cmd)
        self.bs.send(msg.PyProcedureCall(command=cmd))
        r_msg = self.bs.recv(msg.TclProcedureResult)
        cmd_idx = int(r_msg.cmd_idx)
        err_code = int(r_msg.err_code)
        
        if err_code:
            self.log("error", f"{r_msg.result}")
            raise TclError(r_msg.result)
        else:
            self.log("retval", r_msg.result)
            return TclRemoteObjRef(self, cmd_idx, r_msg.result, cmd)
        

class TclToolInterface:
    """
    Entering a with block using a TclTool yields a TclToolInterface.

    Example::

        with MyTclTool() as t:
            # t is a TclToolInterface object.
            pass
    """
    def __init__(self, tcl_tool:TclTool):
        self.tcl_tool = tcl_tool

    def __getattr__(self, name):
        """
        For every unknown attribute, a method is returned that that calls the 
        Tcl command or procedure of corresponding name.

        Example use for a TclToolInterface t, in which the method returned by
        __getattr__ is immediately called, returning a TclRemoteObjRef::

            v = t.expr(3, "*", 5)
        """
        return lambda *args, **kwargs: self.tcl_tool.proc_call(name, args, kwargs)

    def __call__(self, cmd: str) -> TclRemoteObjRef:
        """
        Calling TclToolInterface allows low-level evaluation of a strings without
        using the higher-level method calls.

        Example use for a TclToolInterface t::

            v = t("expr 3 * 5")
        """
        return self.tcl_tool.eval(cmd)



# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

from ..tclobj import encode

def test_obj1():
    o = encode(["hello", "world"])
    assert o == "{{hello} {world}}"

    o = encode("just a string")
    assert o  == "{just a string}"

    o = encode({"a": "b", "c":"d"})
    assert o == "{{a} {b} {c} {d}}"

    o = encode([1,2,3,"abc", "def"])
    assert o == "{{1} {2} {3} {abc} {def}}"


# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

import re
from base64 import b64encode, b64decode
import socket

class RawMessage(dict):
    """Keys can only contain a-z, A-Z and underscores.
    RawMessages can be decoded to Messages."""
    @classmethod
    def from_pipe(cls, f_in):
        """reads until EOF"""
        msg_list = f_in.read().split(b"\n")
        if len(msg_list)%2 != 0:
            raise ValueError("RawMessage requires an even number of items (key value pairs)")
    
        msg = cls()
        while len(msg_list)>0:
            key = msg_list.pop(0).decode("ascii")
            value = msg_list.pop(0)

            msg[key] = b64decode(value).decode("utf8")

        return msg


    def send_to_pipe(self, f_out):
        msg_list = []
        for key, value in self.items():
            if not re.fullmatch("[a-zA-Z_+]*", key):
                raise ValueError(f"Messages keys can only contain a-z, A-Z and underscores, got '{key}'")
            msg_list.append(key.encode("ascii"))
            msg_list.append(b64encode(value.encode("utf8")))
        f_out.write(b"\n".join(msg_list))
        f_out.flush()

    def decode(self, msg_classes):
        pass

    def to_message(self, permitted_msg_classes):
        """Converts RawMessage to Message.
        permitted_msg_classes can be either one message class or a list of message classes."""
        if isinstance(permitted_msg_classes, type):
            permitted_msg_classes = (permitted_msg_classes, )

        for cls in permitted_msg_classes:
            try:
                return cls(self)
            except WrongMessageClass:
                pass
        raise WrongMessageClass()

class WrongMessageClass(Exception):
    pass

class Message:
    """Overwrite keys_optional and keys_required in subclasses."""
    keys_optional = []
    keys_required = []

    __slots__=("_data",)

    def __init__(self, source: dict=None, **kwargs):
        """Can either be initialized from dict that has been received or by providing values for all fields."""
        if source:
            assert len(kwargs)==0
            assert not ("class" in kwargs) 
            if source["class"] != self.__class__.__name__:
                raise WrongMessageClass()
            self._data = dict(source)
        else:
            self._data = dict(kwargs)
            self._data["class"] = self.__class__.__name__

    def to_raw_message(self):
        return RawMessage(self._data)

    def __repr__(self):
        cls_name = self.__class__.__name__
        return f"<{cls_name} {self._data}>"

    def __getattr__(self, name):
        return self._data[name]


# SPDX-FileCopyrightText: 2024 Tobias Kaiser <mail@tb-kaiser.de>
# SPDX-License-Identifier: Apache-2.0

import pytest
import shutil
from .. import TclTool, TclError

class CustomTclsh(TclTool):
    tclsh_name = "test-notcl-tclsh"

    def cmdline(self):
        return [self.tclsh_name, self.script_name()]

    @classmethod
    def exists(cls):
        return shutil.which(cls.tclsh_name) != None

    @classmethod
    def require(cls):
        if not cls.exists():
            pytest.skip("test-notcl-tclsh, which is required for test, was not found in PATH.")

class CustomTclshObjposLast(CustomTclsh):
    called_object_pos = "last"

class CustomTclshObjposFirst(CustomTclsh):
    called_object_pos = "first"

def test_expr():
    CustomTclsh.require()

    with CustomTclsh() as t:
        v=t.expr(9, '+', 3, '*', 11)
        assert int(v) == 9+3*11

def test_box_ref():
    CustomTclsh.require()
    
    with CustomTclsh() as t:
        mybox = t.create_box(999)
        v = t.unwrap_box(mybox)
        assert int(v) == 999

def test_box_ref_by_string():
    CustomTclsh.require()
    
    with CustomTclsh() as t:
        mybox = t.create_box(5432)
        mybox_str = str(mybox)
        with pytest.raises(TclError):
            t.unwrap_box(mybox_str)

        v = t.unwrap_box(mybox)
        assert int(v)==5432


def test_box_ref_multiple():
    CustomTclsh.require()
    
    with CustomTclsh() as t:
        box1 = t.create_box(1234)
        box2 = t.create_box(5678)
        box3 = t.create_box(9012)
        v = t.unwrap_box(box1)
        assert int(v) == 1234
        v = t.unwrap_box(box3)
        assert int(v) == 9012
        v = t.unwrap_box(box2)
        assert int(v) == 5678

def test_box_as_object():
    CustomTclsh.require()

    assert CustomTclsh.called_object_pos == 'second'
    with CustomTclsh() as t:
        box1 = t.create_box(1234)
        v = box1.unwrap_box()
        assert int(v) == 1234

    assert CustomTclshObjposLast.called_object_pos == 'last'
    with CustomTclshObjposLast() as t:
        box1 = t.create_box(1234)
        v = box1.unwrap_box()
        assert int(v) == 1234

def test_called_object_pos():
    CustomTclsh.require()

    assert CustomTclsh.called_object_pos == 'second'
    with CustomTclsh() as t:
        l = t.list("a b c d")
        v = l.list('hello', 'world', whats="up")
        assert str(v) == "-whats up {{a b c d}} hello world"

    assert CustomTclshObjposLast.called_object_pos == 'last'
    with CustomTclshObjposLast() as t:
        l = t.list("a b c d")
        v = l.list('hello', 'world', whats="up")
        assert str(v) == "-whats up hello world {{a b c d}}"

    assert CustomTclshObjposFirst.called_object_pos == 'first'
    with CustomTclshObjposFirst() as t:
        # Here we generate a "object" that is just the built-in list command.
        l = t.list("list")
        v = l.mycmd('hello', 'world', whats="up")
        assert str(v) == '-whats up mycmd hello world'


def test_cmdline():
    CustomTclsh.require()

    t_proto = CustomTclsh()

    assert t_proto.cmdline() == ['test-notcl-tclsh', '<placeholder for notcl.tcl>']
    with CustomTclsh() as t:
        assert t_proto.cmdline() == ['test-notcl-tclsh', '<placeholder for notcl.tcl>']